import datetime
from typing import Union, Dict
import uuid
import asyncio
from asyncio import AbstractEventLoop, Future, TimeoutError

from .request import RPCMessageType, RPCRequest, RPCResponse, response_from_raw
from ..router import Router, HeaderMethodRouter
from ..serializers import serialize, deserialize, PayloadTypeNotSupportedException, SerializeFailedException,\
    DeserializeFailedException, ContentTypeNotSupportedException
from ..exceptions import AiopikaException
from .exceptions import PublishMessageException, DeliveryException, MessageTimeoutException, ExternalException,\
    ReceiveMessageException, ReplyNotSupportBroadcastException

from aio_pika import connect_robust, Connection, Channel, IncomingMessage, Message, Exchange, Queue, ExchangeType
from aio_pika.message import DeliveredMessage

from logging import getLogger
log = getLogger('AiopikaClient')


DateType = Union[int, datetime.datetime, float, datetime.timedelta, None]


class AsyncResult(object):

    #: Timeout for wait publish new message in broker
    rpc_wait_timeout = 60

    def __init__(self, client, exchange: str, queue: str, identifier: str, correlation_id: str, reply: bool = True):
        self._client = client
        self._exchange = exchange
        self._queue = queue
        self._identifier = identifier
        self._correlation_id = correlation_id
        self._reply = reply

    @property
    def exchange(self) -> str:
        return self._exchange

    @property
    def queue(self) -> str:
        return self._queue

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def correlation_id(self) -> str:
        return self._correlation_id

    @property
    def reply(self) -> bool:
        return self._reply

    async def wait(self, timeout: int = None):
        if not self.reply:
            raise ReplyNotSupportBroadcastException

        if timeout is None:
            timeout = self.rpc_wait_timeout

        return await self._client.wait_result(self, timeout=timeout)


class AiopikaClient(object):

    #: Timeout for wait declaring queue or exchange
    declaring_timeout = 30

    #: Timeout for wait publish new task in broker
    publish_timeout = 30

    #: Timeout for wait publish new task in broker
    rpc_wait_timeout = 60

    #: Limit for count retries task
    max_retries = 100

    def __init__(self, name: str = 'client', host: str = 'localhost',
                 port: int = 5672, virtual_host: str = '/',
                 exchange: str = '', user: str = None,
                 password: str = None, loop: AbstractEventLoop = None):
        self._name = name
        self._host = host
        self._virtual_host = virtual_host
        self._exchange_name = exchange
        self._port = port
        self._user = user
        self._password = password
        self._loop = loop

        self._callback_queue: Queue = None
        self._callback_queue_name = None
        self._callback_queue_consumer_tag = None

        self._connection: Connection = None
        self._channel: Channel = None
        self._exchange: Exchange = None
        self._futures: Dict[str, Future] = {}

        self.router_cls: Router = HeaderMethodRouter

    @property
    def name(self) -> str:
        return self._name

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def user(self) -> str:
        return self._user

    @property
    def password(self) -> str:
        return self._password

    @property
    def loop(self) -> AbstractEventLoop:
        return self._loop

    def replace_loop(self, loop: AbstractEventLoop):
        self._loop = loop

    @property
    def connection(self) -> Connection:
        return self._connection

    @property
    def channel(self) -> Channel:
        return self._channel

    @property
    def exchange(self) -> Exchange:
        return self._exchange

    @property
    def is_closed(self) -> bool:
        return self._connection.is_closed if self._connection is not None else True

    def get_nodename(self):
        return self._name

    async def connect(self, exchange_name: str = 'rpc', loop: AbstractEventLoop = None):
        self._loop = loop or self._loop

        await self._reset_callback_queue()

        log.info(f'Connect to {self._host}:{self._port}/{self._virtual_host} ({self._user}:******)')
        self._connection = await connect_robust(
            host=self._host,
            port=self._port,
            login=self._user,
            password=self._password,
            virtualhost=self._virtual_host,
            loop=self.loop
        )
        self._connection.loop
        self._channel = await self._connection.channel()
        self._exchange = await self._channel.declare_exchange(
            name=exchange_name,
            type=ExchangeType.TOPIC,
            durable=True,
            auto_delete=False
        )

        if self._callback_queue is None:
            await self._reset_callback_queue()
            await self._setup_callback_queue()

    async def close(self):
        # if self._callback_queue is not None:
        #     await self._callback_queue.delete()

        if self._channel is not None:
            await self._channel.close()
            self._channel = None

        if self._connection is not None:
            await self._connection.close()
            self._connection = None

    async def _reset_callback_queue(self):
        if self._callback_queue is not None:
            await self._callback_queue.cancel(self._callback_queue_consumer_tag, timeout=self.declaring_timeout)

        self._callback_queue_name = f'{self._name}.clients.{str(uuid.uuid4())}'

    async def _setup_callback_queue(self):
        self._callback_queue = await self.channel.declare_queue(
            name=self._callback_queue_name,
            # exclusive=True,
            auto_delete=True,
            # durable=True,
            timeout=self.declaring_timeout,
        )

        self._callback_queue_consumer_tag = await self._callback_queue.consume(self._on_response)

    async def _on_response(self, message: IncomingMessage):
        async with message.process(ignore_processed=True):
            if message.correlation_id not in self._futures:
                await message.reject(requeue=False)
                return

            future = self._futures.pop(message.correlation_id)

            try:
                response = response_from_raw(message.body, message.content_type, message.type)

                if response.type == RPCMessageType.result:
                    future.set_result(response.payload)
                elif response.type == RPCMessageType.error:
                    future.set_exception(response.payload)
            except (DeserializeFailedException, ContentTypeNotSupportedException) as e:
                future.set_exception(ReceiveMessageException(message.correlation_id, message.routing_key, message.type))

            await message.ack()

    async def _get_message(self, body, content_type: str, type: str, correlation_id: str, identifier: str,
                     reply_to: str = None, expiration: DateType = None) -> Message:
        return Message(
            body=body,
            content_type=content_type,
            type=type,

            correlation_id=correlation_id,
            reply_to=reply_to,

            expiration=expiration,

            headers={
                'lang': 'py',
                'method': identifier,
                # 'id': correlation_id,
                # 'shadow': shadow,
                # 'countdown': countdown,
                # 'eta': eta,
                # 'group': group_id,
                # 'retries': retries,
                # 'timelimit': [time_limit, soft_time_limit],
                # 'root_id': root_id,
                # 'parent_id': parent_id,
                'origin': self.get_nodename()
            }
        )

    async def _publish_message(self, message: Message, identifier: str, correlation_id: str,
                               routing_key: str) -> DeliveredMessage:
        try:
            result: DeliveredMessage = await self._exchange.publish(
                message=message,
                routing_key=routing_key,
                timeout=self.publish_timeout
            )

            # TODO: except all error codes from: https://cwiki.apache.org/confluence/display/qpid/AMQP+Error+Codes#space-menu-link-content
            if isinstance(result, DeliveredMessage):
                raise DeliveryException(result)

            return result
        except TimeoutError as e:
            raise PublishMessageException(routing_key, identifier, correlation_id)

    async def call(self, identifier: str, queue: str, payload=None, reply: bool = True,
                   expires: DateType = None, *args, **kwargs) -> AsyncResult:
        correlation_id = str(uuid.uuid4())
        body, content_type = serialize(payload)

        message = await self._get_message(
            body=body,
            content_type=content_type,
            type=RPCMessageType.call.value,
            correlation_id=correlation_id,
            identifier=identifier,
            reply_to=self._callback_queue_name if reply else None,
            expiration=expires
        )

        await self._publish_message(
            message=message,
            identifier=identifier,
            correlation_id=correlation_id,
            routing_key=queue
        )

        return AsyncResult(
            self,
            self._channel.default_exchange.name,
            queue,
            identifier,
            correlation_id,
            reply=reply
        )

    async def wait_result(self, result: AsyncResult, timeout: int = None):
        if self._callback_queue is None:
            await self._reset_callback_queue()
            await self._setup_callback_queue()

        if timeout is None:
            timeout = self.rpc_wait_timeout

        future = self.loop.create_future()
        self._futures[result.correlation_id] = future

        try:
            return await asyncio.wait_for(future, timeout=timeout, loop=self.loop)
        except TimeoutError as e:
            raise MessageTimeoutException(result.queue, result.identifier, result.correlation_id)
        except AiopikaException as e:
            raise
        except Exception as e:
            raise


from sentry_sdk import capture_exception

from .result import AiopikaResult

from macrobase_driver.endpoint import Endpoint

from aio_pika import IncomingMessage

from structlog import get_logger
log = get_logger('AiopikaEndpoint')


class AiopikaEndpoint(Endpoint):
    """
    Endpoint class for process incoming messages and ack/nack/reject message
    """

    async def handle(self, driver, message: IncomingMessage, data, *args, **kwargs) -> AiopikaResult:
        """
        Handle method for process incoming message with identifier

        Args:
            driver: Aiopika Macrobase driver
            message (IncomingMessage): Incoming message from driver processing
            data: Deserialized payload from Incoming Message
            *args: Additional arguments
            **kwargs: Additional arguments with keys

        Returns:
            AiopikaResult: Aiopika result action or None  (if return None then driver ack message).
        """
        try:
            result = await self.method(driver, message, data, *args, **kwargs)
        except Exception as e:
            capture_exception(e)
            raise

        return result if result is not None else AiopikaResult()

    async def method(self, driver, message: IncomingMessage, data, *args, **kwargs) -> AiopikaResult:
        return AiopikaResult()


class HealthEndpoint(AiopikaEndpoint):

    async def method(self, driver, message: IncomingMessage, data, *args, **kwargs):
        log.info('Health')
        return AiopikaResult(payload={'status': 'health', 'value': 'ok'})



class AiopikaException(Exception):
    requeue = None


class RoutingException(AiopikaException):
    pass


class PayloadTypeNotSupportedException(AiopikaException):
    requeue = False


class ContentTypeNotSupportedException(AiopikaException):
    requeue = False


class MethodNotFoundException(AiopikaException):
    requeue = False


class EndpointNotImplementedException(AiopikaException):
    pass


class ResponseFailedSendException(AiopikaException):
    pass


class SerializeFailedException(AiopikaException):
    requeue = False


class DeserializeFailedException(AiopikaException):
    requeue = False


class IncomingRoutingFailedException(AiopikaException):
    pass


class ResultDeliveryFailedException(AiopikaException):
    requeue = False



from macrobase_driver.config import BaseConfig, DriverConfig, fields


class RabbitmqPropertyConfig(BaseConfig):
    host = fields.Str('localhost')
    port = fields.Int(5672)

    user        = fields.Str('rabbitmq')
    password    = fields.Str('test')

    vhost = fields.Str('/')


class QueuePropertyConfig(BaseConfig):
    name        = fields.Str('queue')
    auto_delete = fields.Bool(False)
    durable     = fields.Bool(True)


class AiopikaDriverConfig(DriverConfig):

    logo = fields.Str("""
 _____       _
|  __ \     (_)               
| |  | |_ __ ___   _____ _ __ 
| |  | | '__| \ \ / / _ \ '__|
| |__| | |  | |\ V /  __/ |   
|_____/|_|  |_| \_/ \___|_|aiopika""")

    workers = fields.Int(1)

    health_endpoint = fields.Bool(False)

    rabbitmq: RabbitmqPropertyConfig    = fields.Nested(RabbitmqPropertyConfig)
    queue: QueuePropertyConfig          = fields.Nested(QueuePropertyConfig)

    # Processing
    prefetch_count      = fields.Int(10)
    ignore_processed    = fields.Bool(True)
    requeue_delay       = fields.Int(10)
    default_retry_delay = fields.Int(60)
    requeue_unknown     = fields.Bool(False)
    requeue_if_failed   = fields.Bool(True)  # TODO: Set `requeue` for all AiopikaException subclasses

    sentry_dsn = fields.Str('', env_key='SENTRY_DSN')
    sentry_env = fields.Str('', env_key='SENTRY_ENV')



from .driver import AiopikaRPCDriver
from .config import AiopikaRPCDriverConfig
from .request import RPCRequest, RPCResponse, RPCMessageType
from .endpoint import AiopikaRPCEndpoint
from .client import AiopikaClient, AsyncResult



from enum import Enum, IntEnum

from .serializers import serialize, SerializeFailedException, PayloadTypeNotSupportedException

from aio_pika.message import Message, DateType, DeliveryMode

from structlog import get_logger
log = get_logger('AiopikaResult')


class AiopikaResultAction(IntEnum):
    nothing = 0
    ack = 10
    nack = 20
    reject = 30


class AiopikaResult(object):
    """
    Result class with data for action on IncomingMessage
    """

    def __init__(self, action: AiopikaResultAction = AiopikaResultAction.ack,
                 multiple: bool = False, requeue: bool = False,
                 payload=None, headers: dict = None,
                 content_type: str = None, content_encoding: str = None,
                 delivery_mode: DeliveryMode = None,
                 priority: int = None, correlation_id=None,
                 reply_to: str = None, expiration: DateType = None,
                 message_id: str = None,
                 timestamp: DateType = None,
                 type: str = None, user_id: str = None,
                 app_id: str = None):
        """
        Initialize AiopikaResult

        Args:
            action (AiopikaResultAction): Action for IncomingMessage
            multiple (bool): Multiple actio, only for ack/nack actions
            requeue (bool): Requeue flag, only for nack/reject actions

            payload: Payload data object or dict or any supported serialization types (show on serializers.py)
            headers (dict): Headers dictionary for additional information in result message
            content_type (str): Type of body (content) result message
            content_encoding (str): Encoding of body (content) result message
            delivery_mode (DeliveryMode): Delivery mode
            priority (int): Priority of delivery result message
            correlation_id: If of correlation for find result among all others
            reply_to (str): Name of queue for reply
            expiration (DateType): Time for expire result message
            message_id  (str): Id of message
            timestamp (DateType): Timestamp of result message
            type (str): Type of result message
            user_id (str): Id of user for result message
            app_id (str): Id of application for result message
        """
        self.action = action
        self.multiple = multiple
        self.requeue = requeue

        self.payload = payload
        self.headers: dict = headers
        self.content_type: str = content_type
        self.content_encoding: str = content_encoding
        self.delivery_mode: DeliveryMode = delivery_mode
        self.priority: int = priority
        self.correlation_id = correlation_id
        self.reply_to: str = reply_to
        self.expiration: DateType = expiration
        self.message_id: str = message_id
        self.timestamp: DateType = timestamp
        self.type: str = type
        self.user_id: str = user_id
        self.app_id: str = app_id

    def get_response_message(self, payload=None, headers: dict = None,
                             content_type: str = None, content_encoding: str = None,
                             delivery_mode: DeliveryMode = None,
                             priority: int = None, correlation_id=None,
                             reply_to: str = None, expiration: DateType = None,
                             message_id: str = None,
                             timestamp: DateType = None,
                             type: str = None, user_id: str = None,
                             app_id: str = None) -> Message:
        # body = ''.encode()
        # _content_type = 'text/plain'

        # try:
        body, _content_type = serialize(payload or self.payload)
        # except (PayloadTypeNotSupportedException, SerializeFailedException) as e:
        #     pass

        _headers = self.headers

        if headers is not None:
            if _headers is None:
                _headers = headers
            else:
                _headers.update(headers)

        return Message(
            body,
            content_type=content_type or _content_type or self.content_type,
            content_encoding=content_encoding or self.content_encoding,

            headers=_headers,
            delivery_mode=delivery_mode or self.delivery_mode,
            priority=priority or self.priority,
            correlation_id=correlation_id or self.correlation_id,
            reply_to=reply_to or self.reply_to,
            expiration=expiration or self.expiration,
            message_id=message_id or self.message_id,
            timestamp=timestamp or self.timestamp,
            type=type or self.type,
            user_id=user_id or self.user_id,
            app_id=app_id or self.app_id
        )



from abc import ABC, abstractmethod
from typing import List, Dict, Type, Tuple

from .exceptions import SerializeFailedException, DeserializeFailedException, ContentTypeNotSupportedException, \
    PayloadTypeNotSupportedException

import pickle
import rapidjson


class Serializer(ABC):
    input_type = None
    content_type = None

    @staticmethod
    @abstractmethod
    def serialize(value) -> bytes:
        pass

    @staticmethod
    @abstractmethod
    def deserialize(raw: bytes):
        pass


class StringSerializer(Serializer):
    input_type = str
    content_type = 'text/plain'

    @staticmethod
    def serialize(value: str) -> bytes:
        return value.encode()

    @staticmethod
    def deserialize(raw: bytes) -> str:
        return raw.decode()


class JSONSerializer(Serializer):
    input_type = dict
    content_type = 'application/json'

    @staticmethod
    def serialize(value: dict) -> bytes:
        return rapidjson.dumps(value).encode()

    @staticmethod
    def deserialize(raw: bytes) -> dict:
        return rapidjson.loads(raw)


class ExceptionSerializer(Serializer):
    input_type = Exception
    content_type = 'application/python-pickle'

    @staticmethod
    def serialize(value: Exception) -> bytes:
        return pickle.dumps(value)

    @staticmethod
    def deserialize(raw: bytes) -> Exception:
        return pickle.loads(raw)


_serializers: List[Serializer] = [StringSerializer, JSONSerializer, ExceptionSerializer]
deserializers: Dict[str, Serializer] = {s.content_type: s for s in _serializers}


def serialize(value) -> Tuple[bytes, str]:
    value = value or ''
    
    serializer = [s for s in _serializers if isinstance(value, s.input_type)]
    serializer = serializer[0] if serializer else None

    if serializer is None:
        # TODO: add more info about value (short desc) and his type
        raise PayloadTypeNotSupportedException

    try:
        return serializer.serialize(value), serializer.content_type
    except Exception as e:
        raise SerializeFailedException


def deserialize(raw: bytes, content_type: str):
    serializer = deserializers.get(content_type)

    if serializer is None:
        # TODO: add more info about value (short desc) and his type
        raise ContentTypeNotSupportedException

    try:
        return serializer.deserialize(raw)
    except Exception as e:
        raise DeserializeFailedException



from ..exceptions import AiopikaException

from aio_pika.message import DeliveredMessage


class AiopikaRPCException(AiopikaException):
    requeue = None


class PublishMessageException(AiopikaRPCException):

    def __init__(self, queue: str, task: str, correlation_id: str):
        super(PublishMessageException, self).__init__(f'<Queue: {queue} task: {task} correlation_id: {correlation_id}> Publish task error')


class DeliveryException(AiopikaRPCException):

    def __init__(self, message: DeliveredMessage):
        self.exchange = message.delivery.exchange
        self.routing_key = message.delivery.routing_key
        self.reply_code = message.delivery.reply_code
        self.reply_text = message.delivery.reply_text

        super(DeliveryException, self).__init__(
            f'<Exchange: {self.exchange} routing_key: {self.routing_key} code: {self.reply_code}> {self.reply_text}'
        )


class ReceiveMessageException(AiopikaRPCException):

    def __init__(self, correlation_id: str, routing_key: str, type: str):
        super(ReceiveMessageException, self).__init__(f'<Message: {correlation_id} routing_key: {routing_key} type: {type}> Message fail receive')


class MessageTimeoutException(AiopikaRPCException):

    def __init__(self, queue: str, task: str, correlation_id: str):
        super(MessageTimeoutException, self).__init__(f'<Queue: {queue} task: {task} correlation_id: {correlation_id}> Service is unavailable')


class ExternalException(AiopikaRPCException):
    """
    Exception against rpc-fragile
    """

    def __init__(self, exception: Exception):
        super(ExternalException, self).__init__(str(exception.__repr__()))


class ResponseContentException(AiopikaRPCException):
    pass
    # def __init__(self, correlation_id: str, routing_key: str):
    #     super(ResponseContentException, self).__init__(f'<Message: {correlation_id} routing_key: {routing_key}> Message response not serialized')


class ReplyNotSupportBroadcastException(AiopikaRPCException):
    pass



from enum import Enum
import uuid

from ..result import AiopikaResult, AiopikaResultAction, DateType
from ..serializers import deserialize, DeserializeFailedException, ContentTypeNotSupportedException

from aio_pika import Message, IncomingMessage


class RPCMessageType(Enum):
    call = 'call'
    error = 'error'
    result = 'result'


class RPCRequest(object):

    def __init__(self, message: IncomingMessage, identifier: str, payload = None):
        self._message: IncomingMessage = message
        self._identifier: str = identifier
        self._payload = payload

    @property
    def message(self) -> IncomingMessage:
        return self._message

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def payload(self):
        return self._payload


class RPCResponse(object):

    def __init__(self, payload=None, type: RPCMessageType = RPCMessageType.result):
        self.payload = payload
        self.type = type

    def get_result(self, correlation_id: str, identifier: str, expire: DateType) -> AiopikaResult:
        return AiopikaResult(
            action=AiopikaResultAction.ack,
            payload=self.payload,
            correlation_id=correlation_id,
            headers={
                'method': identifier
            },
            expiration=expire,
            type=self.type.value,
        )


def response_from_raw(body: bytes, content_type: str, type_value: str) -> RPCResponse:
    payload = deserialize(body, content_type)

    return RPCResponse(payload=payload, type=RPCMessageType(type_value))



from .endpoint import AiopikaEndpoint
from .exceptions import RoutingException


class Method(object):
    """
    Class for storage handler endpoints and his identifier. This is analog of HTTP Route class.
    """

    def __init__(self, handler: AiopikaEndpoint, identifier: str):
        """
        Initialize of Method

        Args:
            handler (AiopikaEndpoint): Instance of AiopikaEndpoint
            identifier (str): Identifier of method
        """
        super(Method, self).__init__()

        if not isinstance(handler, AiopikaEndpoint):
            raise RoutingException('Handler must be instance of AiopikaEndpoint class')

        self._handler = handler
        self._identifier = identifier

    @property
    def handler(self) -> AiopikaEndpoint:
        return self._handler

    @property
    def identifier(self) -> str:
        return self._identifier



import asyncio
from .request import RPCResponse, RPCMessageType
from .exceptions import ResponseContentException

from ..driver import AiopikaDriver, IncomingMessage
from ..method import Method
from ..router import HeaderMethodRouter, IncomingRoutingFailedException
from ..result import AiopikaResult, AiopikaResultAction
from ..exceptions import PayloadTypeNotSupportedException, SerializeFailedException, ResultDeliveryFailedException, \
    MethodNotFoundException

from structlog import get_logger
log = get_logger('AiopikaDriver')


class AiopikaRPCDriver(AiopikaDriver):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.name = 'Aiopika RPC Driver'
        self.router_cls = HeaderMethodRouter

    async def _process_message(self, message: IncomingMessage):
        method = None

        try:
            method = self._router.get_method(message)
            result = await self._get_method_result(message, method)
            ignore_reply = False
        except MethodNotFoundException as e:
            log.debug(f'Ignore unknown method')
            result = AiopikaResult(action=AiopikaResultAction.nack, requeue=self.config.driver.requeue_unknown)
            ignore_reply = True
        except Exception as e:
            log.error(e)
            result = RPCResponse(payload=e, type=RPCMessageType.error).get_result(message.correlation_id, method.identifier if method is not None else '', message.expiration)
            ignore_reply = True

        await self._process_result(message, result, ignore_reply=ignore_reply)

    async def _process_result(self, message: IncomingMessage, result: AiopikaResult, ignore_reply: bool = False):
        if result.requeue:
            await asyncio.sleep(self.config.driver.requeue_delay)

        if result.action == AiopikaResultAction.ack:
            await message.ack(multiple=result.multiple)
        elif result.action == AiopikaResultAction.nack:
            await message.nack(multiple=result.multiple, requeue=result.requeue)
        elif result.action == AiopikaResultAction.reject:
            await message.reject(requeue=result.requeue)

        if ignore_reply:
            return

        if message.reply_to is not None and len(message.reply_to) != 0:
            try:
                try:
                    result_message = result.get_response_message()
                except (PayloadTypeNotSupportedException, SerializeFailedException) as e:
                    result_message = result.get_response_message(
                        payload=ResponseContentException(),
                        type=RPCMessageType.error.value
                    )

                await self._channel.default_exchange.publish(
                    result_message,
                    routing_key=message.reply_to
                )
            except Exception as e:
                raise ResultDeliveryFailedException

    async def _add_health_if_needed(self):
        if self.config.driver.health_endpoint:
            from .endpoint import HealthEndpoint
            self.add_method(Method(HealthEndpoint(self.context, self.config), 'health'))



import os
import asyncio
import logging.config

from signal import SIG_IGN, SIGINT, SIGTERM, Signals
from signal import signal as signal_func

from typing import List, Dict, Type, Callable, Awaitable

import sentry_sdk
from macrobase_driver.driver import MacrobaseDriver
from macrobase_driver.config import CommonConfig, AppConfig
from macrobase_driver.hook import HookHandler
from macrobase_driver.logging import get_logging_config

from .config import AiopikaDriverConfig
from .hook import AiopikaHookNames
from .result import AiopikaResult, AiopikaResultAction
from .method import Method
from .router import Router, HeaderMethodRouter
from .serializers import deserialize
from .exceptions import AiopikaException, DeserializeFailedException, ContentTypeNotSupportedException,\
    ResultDeliveryFailedException, MethodNotFoundException

import uvloop
from aio_pika import connect_robust, Connection, IncomingMessage, Channel, Queue

from structlog import get_logger

log = get_logger('AiopikaDriver')


class AiopikaDriver(MacrobaseDriver):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.name is None:
            self.name = 'Aiopika Driver'

        self._connection = None
        self._channel: Channel = None
        self._queue: Queue = None

        self._hooks: Dict[AiopikaHookNames, List[HookHandler]] = {}
        self._methods: Dict[str, Method] = {}

        self._router: Router = None
        self.router_cls: Type[Router] = HeaderMethodRouter

        self._setup_sentry()

    @property
    def config(self) -> CommonConfig[AppConfig, AiopikaDriverConfig]:
        return self._config

    def add_hook(self, name: AiopikaHookNames, handler):
        if name not in self._hooks:
            self._hooks[name] = []

        self._hooks[name].append(HookHandler(self, handler))

    def add_method(self, method: Method):
        self._methods[method.identifier] = method

    def add_methods(self, methods: List[Method]):
        self._methods.update({method.identifier: method for method in methods})

    def _setup_sentry(self):
        if self.config.driver.sentry_dsn is None:
            return

        sentry_sdk.init(
            dsn=self.config.driver.sentry_dsn,
            environment=self.config.driver.sentry_env
        )

    async def process_message(self, message: IncomingMessage, *args, **kwargs):
        async with message.process(ignore_processed=self.config.driver.ignore_processed):
            await self._process_message(message)

    async def _process_message(self, message: IncomingMessage):
        try:
            method = self._router.get_method(message)
            log.info(f'<IncomingMessage correlation_id: {message.correlation_id} method: {method}> processing')

            result = await self._get_method_result(message, method)
        except MethodNotFoundException as e:
            log.error(f'<IncomingMessage correlation_id: {message.correlation_id} method: {method}> ignore unknown method')
            result = AiopikaResult(action=AiopikaResultAction.nack, requeue=self.config.driver.requeue_unknown)
            await self._process_result(message, result, ignore_reply=True)
            return
        except Exception as e:
            requeue = e.requeue if isinstance(e, AiopikaException) else self.config.driver.requeue_if_failed

            log.error(e)
            result = AiopikaResult(action=AiopikaResultAction.nack, requeue=requeue)
            await self._process_result(message, result, ignore_reply=True)
            return

        await self._process_result(message, result, ignore_reply=False)

    async def _get_method_result(self, message: IncomingMessage, method: Method):
        data = message.body

        try:
            if message.content_type is not None and len(message.content_type) != 0:
                data = deserialize(message.body, message.content_type)
        except ContentTypeNotSupportedException as e:
            pass

        return await method.handler(self, message, data=data, identifier=method.identifier)

    async def _process_result(self, message: IncomingMessage, result: AiopikaResult, ignore_reply: bool = False):
        if result.requeue:
            await asyncio.sleep(self.config.driver.requeue_delay)

        if result.action == AiopikaResultAction.ack:
            await message.ack(multiple=result.multiple)
        elif result.action == AiopikaResultAction.nack:
            await message.nack(multiple=result.multiple, requeue=result.requeue)
        elif result.action == AiopikaResultAction.reject:
            await message.reject(requeue=result.requeue)

        if ignore_reply:
            return

        if message.reply_to is not None and len(message.reply_to) != 0:
            try:
                result_message = result.get_response_message()

                await self._channel.default_exchange.publish(
                    result_message,
                    routing_key=message.reply_to
                )
            except Exception as e:
                raise ResultDeliveryFailedException

    async def _prepare(self):
        log.debug(f'Router <{self.router_cls.__name__}> initialize')
        self._router = self.router_cls(self._methods)

        self._logging_config = get_logging_config(self.config.app)
        logging.config.dictConfig(self._logging_config)

        await self._add_health_if_needed()

    async def _add_health_if_needed(self):
        if self.config.driver.health_endpoint:
            from .endpoint import HealthEndpoint
            self.add_method(Method(HealthEndpoint(self.context, self.config), 'health'))

    async def _consume(self) -> Connection:
        host            = self.config.driver.rabbitmq.host
        port            = self.config.driver.rabbitmq.port
        user            = self.config.driver.rabbitmq.user
        password        = self.config.driver.rabbitmq.password
        virtual_host    = self.config.driver.rabbitmq.vhost
        queue           = self.config.driver.queue.name

        log.info(f'<Aiopika worker: {os.getpid()}> Connect to {host}:{port}/{virtual_host} ({user}:******)')
        self._connection = await connect_robust(
            host=host,
            port=port,
            login=user,
            password=password,
            virtualhost=virtual_host,

            loop=self.loop
        )

        self._channel = await self._connection.channel()
        await self._channel.set_qos(prefetch_count=self.config.driver.prefetch_count)

        self._queue = await self._channel.declare_queue(
            queue,
            durable=self.config.driver.queue.durable,
            auto_delete=self.config.driver.queue.auto_delete
        )

        await self._queue.consume(self.process_message)

        return self._connection

    def _run_single_mode(self, run_multiple: bool = False, *args, **kwargs):
        uvloop.install()

        pid = os.getpid()
        log.info(f'<Aiopika worker: {pid}> Starting worker')

        await_func = self.loop.run_until_complete

        await_func(self._call_hooks(AiopikaHookNames.before_server_start.value))

        connection: Connection = None

        try:
            connection = await_func(self._consume())

            # Ignore SIGINT when run_multiple
            if run_multiple:
                signal_func(SIGINT, SIG_IGN)

            # Register signals for graceful termination
            _singals = (SIGTERM,) if run_multiple else (SIGINT, SIGTERM)
            for _signal in _singals:
                try:
                    def fu():
                        print('stop')
                        self.loop.stop()

                    self.loop.add_signal_handler(_signal, self.loop.stop)
                except NotImplementedError:
                    log.warning(
                        "AiopikaDriver tried to use loop.add_signal_handler "
                        "but it is not implemented on this platform."
                    )

            self.loop.run_forever()
        except BaseException as e:
            log.error(e)

            if self._connection is not None:
                await_func(connection.close())
        finally:
            if self._connection is not None:
                await_func(self._connection.close())
            log.info(f'<Aiopika worker: {pid}> Stopping worker')

            await_func(self._call_hooks(AiopikaHookNames.after_server_stop.value))
            self.loop.close()

    def _run_multiple_mode(self, workers: int, *args, **kwargs):
        import multiprocessing

        log.debug(self.config.driver.logo)

        processes = []

        def sig_handler(signal, frame):
            log.info("Received signal %s. Shutting down.", Signals(signal).name)
            for process in processes:
                os.kill(process.pid, SIGTERM)

        signal_func(SIGINT, lambda s, f: sig_handler(s, f))
        signal_func(SIGTERM, lambda s, f: sig_handler(s, f))

        for _ in range(workers):
            p = multiprocessing.Process(
                target=self._run_single_mode,
                daemon=True,
                kwargs={
                    'run_multiple': True,
                }
            )
            p.start()

            processes.append(p)

        for process in processes:
            process.join()

        # the above processes will block this until they're stopped
        for process in processes:
            process.terminate()

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        log.debug(self.config.driver.logo)

        self.loop.run_until_complete(self._prepare())

        if self.config.driver.workers == 1:
            self._run_single_mode(*args, **kwargs)
        else:
            self._run_multiple_mode(self.config.driver.workers, *args, **kwargs)



from abc import ABC, abstractmethod
from typing import Dict

from .method import Method
from .exceptions import IncomingRoutingFailedException, MethodNotFoundException

from aio_pika import IncomingMessage


class Router(ABC):
    """
    Abstract class for routing incoming message by any basis.
    """

    def __init__(self, methods: Dict[str, Method]):
        self._methods: Dict[str, Method] = methods

    def get_method(self, message: IncomingMessage) -> Method:
        """
        Route by message

        Args:
            message (IncomingMessage): Incoming message from driver processing

        Returns:
            Method: Routed method
        """
        method = self.get_method_identifier(message)

        if method is None:
            raise IncomingRoutingFailedException

        if method not in self._methods:
            raise MethodNotFoundException

        return self._methods[method]

    @staticmethod
    @abstractmethod
    def get_method_identifier(message: IncomingMessage) -> str:
        """
        Function for getting method identifier from incoming message

        Args:
            message (IncomingMessage): Incoming message from driver processing

        Returns:
            str: Found method identifier or None
        """
        pass


class HeaderMethodRouter(Router):
    """
    The router by method identifier in `method` header of message.
    """

    @staticmethod
    def get_method_identifier(message: IncomingMessage) -> str:
        return message.headers.get('method', None)



from ..config import AiopikaDriverConfig, QueuePropertyConfig


class QueueRPCPropertyConfig(QueuePropertyConfig):
    name: str = 'rpc_queue'
    auto_delete: bool = False
    durable: bool = True


class AiopikaRPCDriverConfig(AiopikaDriverConfig):

    logo: str = """
 _____       _
|  __ \     (_)               
| |  | |_ __ ___   _____ _ __ 
| |  | | '__| \ \ / / _ \ '__|
| |__| | |  | |\ V /  __/ |   
|_____/|_|  |_| \_/ \___|_|aiopika_rpc
"""

    queue: QueueRPCPropertyConfig = QueueRPCPropertyConfig()

    # Processing
    default_expire_call: int = 30



from setuptools import setup, find_packages

setup(
    name='aiopika_macrobase',
    version='2.2.0',
    packages=find_packages(),
    url='https://github.com/mbcores/aiopika-macrobase',
    license='MIT',
    author='Alexey Shagaleev',
    author_email='alexey.shagaleev@yandex.ru',
    description='Aio-pika driver for macrobase framework',
    install_requires=[
        'macrobase-driver>=2.0.0,<3.0.0',
        'aio-pika==5.5.3',
        'uvloop==0.14.0',
        'python-rapidjson==0.8.0',
        'structlog==19.2.0',
        'sentry-sdk>=0.14.3',
        'aiocontextvars==0.2.2'
    ]
)



from sentry_sdk import capture_exception

from .request import RPCRequest, RPCResponse, RPCMessageType
from ..endpoint import AiopikaEndpoint
from ..result import AiopikaResult

from aio_pika import IncomingMessage

from structlog import get_logger
log = get_logger('AiopikaRPCEndpoint')


class AiopikaRPCEndpoint(AiopikaEndpoint):
    """
    RPC implementation for RPC processing
    """

    async def handle(self, driver, message: IncomingMessage, data, *args, **kwargs) -> AiopikaResult:
        """
        Handle method for process incoming message

        Args:
            driver: Aiopika Macrobase driver
            message (IncomingMessage): Incoming message from driver processing
            data: Deserialized payload from Incoming Message
            *args: Additional arguments
            **kwargs: Additional arguments with keys

        Returns:
            AiopikaResult: Aiopika result action or None  (if return None then driver ack message).
        """
        identifier = kwargs.get('identifier', None)
        request = RPCRequest(message, identifier, payload=data)

        try:
            response = await self.method(driver, request, request.payload, *args, **kwargs)
        except Exception as e:
            capture_exception(e)
            response = RPCResponse(e, type=RPCMessageType.error)

        return response.get_result(message.correlation_id, identifier, message.expiration)

    async def method(self, driver, request: RPCRequest, data, *args, **kwargs) -> RPCResponse:
        return RPCResponse()


class HealthEndpoint(AiopikaRPCEndpoint):

    async def method(self, driver, request: RPCRequest, data, *args, **kwargs) -> RPCResponse:
        log.info('Health')
        return RPCResponse(payload={'status': 'health', 'value': 'ok'})
